#include "slick.sh"
#import "main.e"
#import "stdprocs.e"
#import "sc/lang/IToString.e"
#import "sc/lang/IHashable.e"
#import "colorDistance.e"
#include "markers.sh"

static void dbgsay(_str msg)
{
    //_SccDisplayOutput("SYM: " msg);
    //say("SYM: " msg);
}

class WordInfo : sc.lang.IToString, sc.lang.IHashable
{
    public _str m_Word;
    int m_Color;
    boolean m_Enabled;
    boolean m_WordMatch;

    WordInfo(_str word = '', int color = -1, boolean wordMatch = true)
    {
        m_Word = word;
        m_Color = color;
        m_Enabled = true;
        m_WordMatch = wordMatch;
    }

    _str toString()
    {
        return m_Word;
    }

    _str toStringDbg()
    {
        return (m_Enabled ? "E":"e") :+
               (m_WordMatch ? "W":"w") :+
               " " m_Color " " :+
               m_Word;
    }

    _str getHashKey()
    {
        return m_Word m_Color m_WordMatch;
    }

    _str getRe()
    {
        if (m_WordMatch)
        {
            return "\\b" _escape_re_chars(m_Word, 'U') "\\b";
        }
        else
        {
            return _escape_re_chars(m_Word, 'U');
        }
    }
};

static int s_modtimeSymHighlight = -1;
static int s_markertypeSymHighlight = -1;
static int s_timerSymHighlight = -1;
static _str s_searchString = null;
static WordInfo s_symTags[];

int def_sym_highlight_delay = 500;
boolean def_sym_highlight_use_scrollmarkers = 1;
boolean def_sym_highlight_confirm_clear = 1;
boolean def_sym_highlight_persist_across_sessions = 0;

static struct SymColour
{
    int m_rgb;
    _str m_name;
    int m_cRef;    // used for delayed color creation and selection.
    int m_ScrollMarkerType;
    int m_ColourSymHighlight;
};

#define RGB(r, g, b) (((b)<<16)|((g)<<8)|(r))

static SymColour s_symColour[] =
{
    { RGB(0xF0, 0xF8, 0xFF), "AliceBlue"},
    { RGB(0xFA, 0xEB, 0xD7), "AntiqueWhite"},
    { RGB(0x7F, 0xFF, 0xD4), "Aquamarine"},
    { RGB(0xF0, 0xFF, 0xFF), "Azure"},
    { RGB(0xF5, 0xF5, 0xDC), "Beige"},
    { RGB(0xFF, 0xE4, 0xC4), "Bisque"},
    { RGB(0x00, 0x00, 0x00), "Black"},
    { RGB(0xFF, 0xEB, 0xCD), "BlanchedAlmond"},
    { RGB(0x00, 0x00, 0xFF), "Blue"},
    { RGB(0x8A, 0x2B, 0xE2), "BlueViolet"},
    { RGB(0xA5, 0x2A, 0x2A), "Brown"},
    { RGB(0xDE, 0xB8, 0x87), "BurlyWood"},
    { RGB(0x5F, 0x9E, 0xA0), "CadetBlue"},
    { RGB(0x7F, 0xFF, 0x00), "Chartreuse"},
    { RGB(0xD2, 0x69, 0x1E), "Chocolate"},
    { RGB(0xFF, 0x7F, 0x50), "Coral"},
    { RGB(0x64, 0x95, 0xED), "CornflowerBlue"},
    { RGB(0xFF, 0xF8, 0xDC), "Cornsilk"},
    { RGB(0xDC, 0x14, 0x3C), "Crimson"},
    { RGB(0x00, 0xFF, 0xFF), "Cyan"},
    { RGB(0x00, 0x00, 0x8B), "DarkBlue"},
    { RGB(0x00, 0x8B, 0x8B), "DarkCyan"},
    { RGB(0xB8, 0x86, 0x0B), "DarkGoldenRod"},
    { RGB(0xA9, 0xA9, 0xA9), "DarkGray"},
    { RGB(0x00, 0x64, 0x00), "DarkGreen"},
    { RGB(0xBD, 0xB7, 0x6B), "DarkKhaki"},
    { RGB(0x8B, 0x00, 0x8B), "DarkMagenta"},
    { RGB(0x55, 0x6B, 0x2F), "DarkOliveGreen"},
    { RGB(0xFF, 0x8C, 0x00), "DarkOrange"},
    { RGB(0x99, 0x32, 0xCC), "DarkOrchid"},
    { RGB(0x8B, 0x00, 0x00), "DarkRed"},
    { RGB(0xE9, 0x96, 0x7A), "DarkSalmon"},
    { RGB(0x8F, 0xBC, 0x8F), "DarkSeaGreen"},
    { RGB(0x48, 0x3D, 0x8B), "DarkSlateBlue"},
    { RGB(0x2F, 0x4F, 0x4F), "DarkSlateGray"},
    { RGB(0x00, 0xCE, 0xD1), "DarkTurquoise"},
    { RGB(0x94, 0x00, 0xD3), "DarkViolet"},
    { RGB(0xFF, 0x14, 0x93), "DeepPink"},
    { RGB(0x00, 0xBF, 0xFF), "DeepSkyBlue"},
    { RGB(0x69, 0x69, 0x69), "DimGray"},
    { RGB(0x1E, 0x90, 0xFF), "DodgerBlue"},
    { RGB(0xB2, 0x22, 0x22), "FireBrick"},
    { RGB(0xFF, 0xFA, 0xF0), "FloralWhite"},
    { RGB(0x22, 0x8B, 0x22), "ForestGreen"},
    { RGB(0xDC, 0xDC, 0xDC), "Gainsboro"},
    { RGB(0xF8, 0xF8, 0xFF), "GhostWhite"},
    { RGB(0xFF, 0xD7, 0x00), "Gold"},
    { RGB(0xDA, 0xA5, 0x20), "GoldenRod"},
    { RGB(0x80, 0x80, 0x80), "Gray"},
    { RGB(0x00, 0x80, 0x00), "Green"},
    { RGB(0xAD, 0xFF, 0x2F), "GreenYellow"},
    { RGB(0xF0, 0xFF, 0xF0), "HoneyDew"},
    { RGB(0xFF, 0x69, 0xB4), "HotPink"},
    { RGB(0xCD, 0x5C, 0x5C), "IndianRed"},
    { RGB(0x4B, 0x00, 0x82), "Indigo"},
    { RGB(0xFF, 0xFF, 0xF0), "Ivory"},
    { RGB(0xF0, 0xE6, 0x8C), "Khaki"},
    { RGB(0xE6, 0xE6, 0xFA), "Lavender"},
    { RGB(0xFF, 0xF0, 0xF5), "LavenderBlush"},
    { RGB(0x7C, 0xFC, 0x00), "LawnGreen"},
    { RGB(0xFF, 0xFA, 0xCD), "LemonChiffon"},
    { RGB(0xAD, 0xD8, 0xE6), "LightBlue"},
    { RGB(0xF0, 0x80, 0x80), "LightCoral"},
    { RGB(0xE0, 0xFF, 0xFF), "LightCyan"},
    { RGB(0xFA, 0xFA, 0xD2), "LightGoldenRodYellow"},
    { RGB(0xD3, 0xD3, 0xD3), "LightGray"},
    { RGB(0x90, 0xEE, 0x90), "LightGreen"},
    { RGB(0xFF, 0xB6, 0xC1), "LightPink"},
    { RGB(0xFF, 0xA0, 0x7A), "LightSalmon"},
    { RGB(0x20, 0xB2, 0xAA), "LightSeaGreen"},
    { RGB(0x87, 0xCE, 0xFA), "LightSkyBlue"},
    { RGB(0x77, 0x88, 0x99), "LightSlateGray"},
    { RGB(0xB0, 0xC4, 0xDE), "LightSteelBlue"},
    { RGB(0xFF, 0xFF, 0xE0), "LightYellow"},
    { RGB(0x00, 0xFF, 0x00), "Lime"},
    { RGB(0x32, 0xCD, 0x32), "LimeGreen"},
    { RGB(0xFA, 0xF0, 0xE6), "Linen"},
    { RGB(0xFF, 0x00, 0xFF), "Magenta"},
    { RGB(0x80, 0x00, 0x00), "Maroon"},
    { RGB(0x66, 0xCD, 0xAA), "MediumAquaMarine"},
    { RGB(0x00, 0x00, 0xCD), "MediumBlue"},
    { RGB(0xBA, 0x55, 0xD3), "MediumOrchid"},
    { RGB(0x93, 0x70, 0xDB), "MediumPurple"},
    { RGB(0x3C, 0xB3, 0x71), "MediumSeaGreen"},
    { RGB(0x7B, 0x68, 0xEE), "MediumSlateBlue"},
    { RGB(0x00, 0xFA, 0x9A), "MediumSpringGreen"},
    { RGB(0x48, 0xD1, 0xCC), "MediumTurquoise"},
    { RGB(0xC7, 0x15, 0x85), "MediumVioletRed"},
    { RGB(0x19, 0x19, 0x70), "MidnightBlue"},
    { RGB(0xF5, 0xFF, 0xFA), "MintCream"},
    { RGB(0xFF, 0xE4, 0xE1), "MistyRose"},
    { RGB(0xFF, 0xE4, 0xB5), "Moccasin"},
    { RGB(0xFF, 0xDE, 0xAD), "NavajoWhite"},
    { RGB(0x00, 0x00, 0x80), "Navy"},
    { RGB(0xFD, 0xF5, 0xE6), "OldLace"},
    { RGB(0x80, 0x80, 0x00), "Olive"},
    { RGB(0x6B, 0x8E, 0x23), "OliveDrab"},
    { RGB(0xFF, 0xA5, 0x00), "Orange"},
    { RGB(0xFF, 0x45, 0x00), "OrangeRed"},
    { RGB(0xDA, 0x70, 0xD6), "Orchid"},
    { RGB(0xEE, 0xE8, 0xAA), "PaleGoldenRod"},
    { RGB(0x98, 0xFB, 0x98), "PaleGreen"},
    { RGB(0xAF, 0xEE, 0xEE), "PaleTurquoise"},
    { RGB(0xDB, 0x70, 0x93), "PaleVioletRed"},
    { RGB(0xFF, 0xEF, 0xD5), "PapayaWhip"},
    { RGB(0xFF, 0xDA, 0xB9), "PeachPuff"},
    { RGB(0xCD, 0x85, 0x3F), "Peru"},
    { RGB(0xFF, 0xC0, 0xCB), "Pink"},
    { RGB(0xDD, 0xA0, 0xDD), "Plum"},
    { RGB(0xB0, 0xE0, 0xE6), "PowderBlue"},
    { RGB(0x80, 0x00, 0x80), "Purple"},
    { RGB(0x66, 0x33, 0x99), "RebeccaPurple"},
    { RGB(0xFF, 0x00, 0x00), "Red"},
    { RGB(0xBC, 0x8F, 0x8F), "RosyBrown"},
    { RGB(0x41, 0x69, 0xE1), "RoyalBlue"},
    { RGB(0x8B, 0x45, 0x13), "SaddleBrown"},
    { RGB(0xFA, 0x80, 0x72), "Salmon"},
    { RGB(0xF4, 0xA4, 0x60), "SandyBrown"},
    { RGB(0x2E, 0x8B, 0x57), "SeaGreen"},
    { RGB(0xFF, 0xF5, 0xEE), "SeaShell"},
    { RGB(0xA0, 0x52, 0x2D), "Sienna"},
    { RGB(0xC0, 0xC0, 0xC0), "Silver"},
    { RGB(0x87, 0xCE, 0xEB), "SkyBlue"},
    { RGB(0x6A, 0x5A, 0xCD), "SlateBlue"},
    { RGB(0x70, 0x80, 0x90), "SlateGray"},
    { RGB(0xFF, 0xFA, 0xFA), "Snow"},
    { RGB(0x00, 0xFF, 0x7F), "SpringGreen"},
    { RGB(0x46, 0x82, 0xB4), "SteelBlue"},
    { RGB(0xD2, 0xB4, 0x8C), "Tan"},
    { RGB(0x00, 0x80, 0x80), "Teal"},
    { RGB(0xD8, 0xBF, 0xD8), "Thistle"},
    { RGB(0xFF, 0x63, 0x47), "Tomato"},
    { RGB(0x40, 0xE0, 0xD0), "Turquoise"},
    { RGB(0xEE, 0x82, 0xEE), "Violet"},
    { RGB(0xF5, 0xDE, 0xB3), "Wheat"},
    { RGB(0xFF, 0xFF, 0xFF), "White"},
    { RGB(0xF5, 0xF5, 0xF5), "WhiteSmoke"},
    { RGB(0xFF, 0xFF, 0x00), "Yellow"},
    { RGB(0x9A, 0xCD, 0x32), "YellowGreen"}
};

static int indexOfWord(WordInfo &sym)
{
    int i;
    for (i = 0; i < s_symTags._length();i++)
    {
        if (s_symTags[i].m_Word == sym.m_Word)
        {
            return i;
        }
    }
    return -1;
}

static int indexOfWordStr(_str &sym)
{
    int i;
    for (i = 0; i < s_symTags._length();i++)
    {
        if (s_symTags[i].m_Word == sym)
        {
            return i;
        }
    }
    return -1;
}

/*-------------------------------------------------------------------------------
    CycleSymHighlight
 
   Advances to next colour for the current word.  If current word isn't already
   highlit, goes to the least used colour
-------------------------------------------------------------------------------*/
_command void CycleSymHighlight() name_info(',')
{
    WordInfo sym = GetCurWordOrSelection();

    // if not highlit, do it.
    int index = indexOfWord(sym);

    if (index == -1)
    {
        ToggleSymHighlight();
    }
    else
    {
        WordInfo *word = &s_symTags[index];

        s_symColour[word->m_Color].m_cRef -= 1;
        word->m_Color = (word->m_Color + 1) % s_symColour._length();
        s_symColour[word->m_Color].m_cRef += 1;

        EnsureColour(word->m_Color);

        UpdateScreen(true /* update now */);
    }
}

WordInfo GetHighlightWords()[]
{
    return s_symTags;
}

SymColour GetSymColor(int index)
{
    if (index < 0 || index > s_symColour._length())
    {
        return null;
    }
    return s_symColour[index];
}

boolean AddHighlightWord(WordInfo &sym)
{
    dbgsay("AddHighlightWord: " sym);
    if (length(sym) != 0)
    {
        int index = indexOfWord(sym);
        if (index == -1)
        {
            boolean result = AddToFullWordTerms(sym);
            UpdateScreen(true /* update now */);
            call_list('_highlight_addword', sym);
            return result;
        }
        else
        {
            sym = s_symTags[index];
            call_list('_highlight_addword', sym);
        }
    }
    return false;
}

boolean RemoveHighlightWord(_str &sym, WordInfo &removed)
{
    dbgsay("RemoveHighlightWord: " sym);
    if (length(sym) != 0)
    {
        int i = indexOfWordStr(sym);
        if (i != -1)
        {
            removed = s_symTags[i];
            s_symTags._deleteel(i);
            s_searchString = null;
            UpdateScreen(true /* update now */);
            return true;
        }
    }
    return false;
}

static void createSearchString()
{
    int len = s_symTags._length();
    s_searchString = '';

    int i;
    for (i = 0; i < len; i++)
    {
        if (s_symTags[0].m_Enabled)
        {
            if (s_searchString == '')
            {
                s_searchString = s_symTags[i].getRe();
            }
            else
            {
                s_searchString :+= '|' :+ s_symTags[i].getRe();
            }
        }
    }
    dbgsay("s_searchString:" s_searchString);
}

static boolean AddToFullWordTerms(WordInfo &sym)
{
    boolean fAdd = false;
    if (length(sym.m_Word) != 0)
    {
        fAdd = true;
        int i = indexOfWord(sym);
        if (i != -1)
        {
            dbgsay("AddToFullWordTerms (removed): " sym);
            fAdd = false;
            s_symTags._deleteel(i);
        }

        if (fAdd)
        {
            int cTags = s_symTags._length();
            s_symTags[cTags] = sym;

            dbgsay("AddToFullWordTerms (added): " sym);
        }

        s_searchString = null;
    }
    return fAdd;
}

static WordInfo GetCurWordOrSelection()
{
    int start_col, end_col;
    WordInfo sym;
    boolean isSelection = false;

    if (_select_type() == "CHAR")
    {
        int dummy;
        _str dummy2;
        int numLines;
        _get_selinfo(start_col, end_col, dummy, '', dummy2, dummy, dummy, numLines);
        if (numLines == 1 && start_col != end_col)
        {
            sym.m_Word = _expand_tabsc(start_col, end_col - start_col, 'S'); // TODO: If tabs are changed to spaces, then the string will never match will it?
            isSelection = true;
            sym.m_WordMatch = false;
        }
    }

    if (!isSelection)
    {
        sym.m_Word = cur_word(start_col);
        sym.m_WordMatch = true;
    }

    return sym;

}

/* T O G G L E  S Y M  H I G H L I G H T */
/*----------------------------------------------------------------------------
    %%Function: ToggleSymHighlight
    %%Author: MarkSun
    %%Owner: MarkSun

   If the IP is on an identifier, either add it or remove it from the list
   of identifiers we're tracking for highlighting purposes.

----------------------------------------------------------------------------*/
_command void ToggleSymHighlight() name_info(',')
{
    int i;
    WordInfo sym = GetCurWordOrSelection();

    dbgsay("ToggleSymHighlight: " sym);
    boolean fAdded;

    fAdded = AddToFullWordTerms(sym);
    if (fAdded)
    {
        call_list('_highlight_addword', sym);
    }
    else
    {
        call_list('_highlight_delword', sym);
    }

    UpdateScreen(true /* update now */);
}

/*-------------------------------------------------------------------------------
    _InitSymColour
-------------------------------------------------------------------------------*/
static void _InitSymColour()
{
    int num = s_symColour._length();
    int i;
    dbgsay("_InitSymColour1");

    if (!def_sym_highlight_persist_across_sessions || s_symColour._length() == 0)
    {
        for(i = 0; i < num; ++i)
        {
            s_symColour[i].m_cRef = 0;
        }
    }
    for(i = 0; i < num; ++i)
    {
        //double delta = getColorDelta(s_symColour[i].m_rgb, 0xffffff);
        //dbgsay("Delta colors " s_symColour[i].m_name ", WHITE == " delta);
        s_symColour[i].m_ColourSymHighlight = -1;
    }

    if (def_sym_highlight_persist_across_sessions || true)
    {
        for (i=0; i < s_symColour._length(); i++)
        {
            //if (s_symColour[i].m_cRef > 0)
            {
                EnsureColour(i);
            }
        }
    }
}


/*-------------------------------------------------------------------------------
    DbgDumpSymHighlight
 
   Dumps out current array values for debugging purposes
-------------------------------------------------------------------------------*/
_command void DbgDumpSymHighlight() name_info(',')
{
    int i;
    dbgsay ("--- SymHighlight dump: " :+ _time('L'));
    dbgsay ("search string:");
    if (s_searchString == null)
    {
        dbgsay ("   " :+ "null");
    }
    else
    {
        dbgsay ("   " :+ s_searchString);
    }
    dbgsay ("word list:");
    for (i = 0; i < s_symTags._length(); i++)
    {
        dbgsay ("   " :+ (i+1) :+ ": " :+ s_symTags[i].toStringDbg() );
    }

    dbgsay ("--- End SymHighlight dump");

}


/*-------------------------------------------------------------------------------
    DoClearAllSymHighlightStructs
-------------------------------------------------------------------------------*/
static void DoClearAllSymHighlightStructs()
{
    int i;

    WordInfo x_symTags[] = s_symTags;
    s_symTags._makeempty();
    s_searchString = '';

    for (i = 0; i < x_symTags._length(); i++)
    {
        call_list('_highlight_delword', x_symTags[i]);
    }

    for (i = 0; i < s_symColour._length(); i++)
    {
        if (s_symColour[i].m_cRef > 0)
        {
            s_symColour[i].m_cRef = 0;
        }
    }
}


/* C L E A R  A L L  S Y M  H I G H L I G H T */
/*----------------------------------------------------------------------------
    %%Function: ClearAllSymHighlight
    %%Author: MarkSun
    %%Owner: MarkSun

   Bring up a dialog listing all identifiers that we are currently highlighting.
   If you respond Yes, then clear the entire list.

----------------------------------------------------------------------------*/
_command void ClearAllSymHighlight() name_info(',')
{
    _str result = IDYES;
    _str messageString = "Clear the following list?\n\n";
    int i;
    int j;

    if (def_sym_highlight_confirm_clear)
    {
        for (i = 0; i < s_symTags._length(); i++)
        {
            messageString :+= (i+1) :+ ": " :+ s_symTags[i] :+ "\n";
        }

        result=_message_box(messageString, '', MB_YESNO|MB_ICONQUESTION);
    }

    if (result==IDYES)
    {
        DoClearAllSymHighlightStructs();
        UpdateScreen(true);
    }
}


/*-------------------------------------------------------------------------------
    RemoveSymHighlightMarkers
-------------------------------------------------------------------------------*/
static void RemoveSymHighlightMarkers(int wid)
{
    int len = s_symColour._length();
    int i;

    for (i = 0; i < len; i++)
    {
        _ScrollMarkupRemoveType(wid, s_symColour[i].m_ScrollMarkerType);
    }
}

/*-------------------------------------------------------------------------------
    RemoveAllSymHighlightMarkerTypes
-------------------------------------------------------------------------------*/
static void RemoveAllSymHighlightMarkerTypes()
{
    int len = s_symColour._length();
    int i;

    for (i = 0; i < len; i++)
    {
        _ScrollMarkupRemoveAllType(s_symColour[i].m_ScrollMarkerType);
    }
}


/*-------------------------------------------------------------------------------
    _UpdateWindow
-------------------------------------------------------------------------------*/
void _UpdateWindow()
{
    typeless p,m,ss,sf,sw,sr,sf2;
    _save_pos2(p);
    save_selection(m);
    save_search(ss, sf, sw, sr, sf2);
    _StreamMarkerRemoveType(p_window_id, s_markertypeSymHighlight);
    RemoveSymHighlightMarkers(p_window_id);

    //dbgsay("Update: " p_buf_name);

    if (s_symTags._length() != 0)
    {
        _str searchArgs;
        searchArgs = '<U@XC';
        _deselect();
        top();

        if (s_searchString == null)
        {
            createSearchString();
        }

        if ( s_searchString != '' && !search(s_searchString, searchArgs) )
        {
            do
            {
                //_str s = strip( get_text( match_length(), (int)_QROffset() ), 'T' );
                _str s = get_text(match_length(), (int)_QROffset());
                long offset_highlight = _QROffset();
                int length_Highlight = s._length();
                int pos_marker = _StreamMarkerAdd( p_window_id, offset_highlight, length_Highlight, true, 0, s_markertypeSymHighlight, '');
                _StreamMarkerSetTextColor(pos_marker, GetSymHighlightColour(s));

                if (def_sym_highlight_use_scrollmarkers)
                {
                    int line=(int)_QLineNumberFromOffset(offset_highlight);
                    _ScrollMarkupAdd(p_window_id, line, GetSymHighlightMarker(s));
                }
            }
            while (!repeat_search(searchArgs));
        }
    }

    restore_search( ss, sf, sw, sr, sf2); 
    restore_selection(m);
    _restore_pos2(p);

    refresh();
}


/* U P D A T E  S C R E E N */
/*----------------------------------------------------------------------------
    %%Function: UpdateScreen
    %%Author: MarkSun
    %%Owner: MarkSun

   Goes through the current buffer and update the highlights based on current
   information.

----------------------------------------------------------------------------*/
static void UpdateScreen(boolean fNow)
{
    get_window_id( auto orig_view);
    activate_window (_mdi.p_child);

    if (s_modtimeSymHighlight != p_LastModified || fNow)
    {
        //dbgsay("UpdateScreen");
        s_modtimeSymHighlight = p_LastModified;

        for_each_mdi_child('-UpdateWindow','');
    }
    else
    {
    }
    activate_window(orig_view); 
}


/* S Y M  H I G H L I G H T  C A L L B A C K */
/*----------------------------------------------------------------------------
    %%Function: SymHighlightCallback
    %%Author: MarkSun
    %%Owner: MarkSun

    Callback to periodically ensure the current buffer has been updated.

----------------------------------------------------------------------------*/
void SymHighlightCallback()
{
    if ( !p_mdi_child || command_state())
    {
        return;
    }
    if (_idle_time_elapsed() < def_sym_highlight_delay)
    {
        return;
    }

    UpdateScreen(false /* update now */);
}


/*-------------------------------------------------------------------------------
    EnsureColour
-------------------------------------------------------------------------------*/
static void EnsureColour(int iColour)
{
    if (s_symColour[iColour].m_ColourSymHighlight == -1)
    {
        s_symColour[iColour].m_ColourSymHighlight = _AllocColor();
//        dbgsay("AllocColor: "s_symColour[iColour].m_ColourSymHighlight); 
        _default_color(s_symColour[iColour].m_ColourSymHighlight, _rgb(0x00, 0x00, 0x00), s_symColour[iColour].m_rgb, 0);
        s_symColour[iColour].m_ScrollMarkerType = _ScrollMarkupAllocType();

        _ScrollMarkupSetTypeColor(s_symColour[iColour].m_ScrollMarkerType, s_symColour[iColour].m_ColourSymHighlight);
    }
}

/*-------------------------------------------------------------------------------
    IGetNextSymColour
-------------------------------------------------------------------------------*/
static int IGetNextSymColour()
{
    int i, iMac;
    int iTarget = 0, iTargetVal = 999;//s_symColour[0].m_cRef;

    iMac = s_symColour._length();

    parse _default_color(CFG_WINDOW_TEXT) with auto fg auto bg auto fontFlags;
    //say("FG: " fg ",  BG: " bg);
    if (iTargetVal != 0)
    {
        for (i = 1; i < iMac; i++)
        {
            double delta1 = getColorDelta(s_symColour[i].m_rgb, (int)fg);
            double delta2 = getColorDelta(s_symColour[i].m_rgb, (int)bg);

            //say("D1: " delta1 ", D2: " delta2);
            if (delta1 >= 2 && delta2 >= 2 && s_symColour[i].m_cRef < iTargetVal)
            {
                iTargetVal = s_symColour[i].m_cRef;
                iTarget = i;
                if (iTargetVal == 0)
                    break;
            }
        }
    }
    s_symColour[iTarget].m_cRef += 1;
    EnsureColour(iTarget);

    if (iTarget < 0 || iTarget >= iMac)
    {
        dbgsay ("Bad colour index returned in IGetNextSymColour: ":+iTarget);
    }
    return iTarget;
}

static void computeDelta(int c1, SymColour &color)
{
    int c2 = color.m_rgb;
    double delta = getColorDelta(c1, c2);
    dbgsay("Color: " color.m_name", C1: " dec2hex(c1) ", C2: " dec2hex(c2) ", Delta="delta);
}

static SymColour *getSymColour(_str &strKey)
{
    int index = indexOfWordStr(strKey);
    if (index == -1)
    {
        dbgsay("WHOOPS! Word not found: " strKey);
        return &s_symColour[0];
    }
    WordInfo *word = &s_symTags[index];

    if (word->m_Color < 0 || word->m_Color >= s_symColour._length())
    {
        //dbgsay ("index for '":+strKey:+"' out of range: ":+word->m_Color);
        word->m_Color = IGetNextSymColour();
        call_list('_highlight_changecolor', s_symTags[index]);
    }

    if (s_symColour[word->m_Color].m_cRef == 0)
    {
        dbgsay ("Hit recovery code in GetSymHighlightColour");
        word->m_Color = IGetNextSymColour();
        call_list('_highlight_changecolor', s_symTags[index]);
    }

    return &s_symColour[word->m_Color];
}

/* G E T  S Y M  H I G H L I G H T  C O L O U R */
/*----------------------------------------------------------------------------
    %%Function: GetSymHighlightColor
    %%Author: Setup the highlighting colour
    %%Owner: Setup the highlighting colour

    

----------------------------------------------------------------------------*/
static int GetSymHighlightColour(_str &strKey)
{
    SymColour *symColor = getSymColour(strKey);
    return symColor->m_ColourSymHighlight;
}

/*-------------------------------------------------------------------------------
    GetSymHighlightMarker
-------------------------------------------------------------------------------*/
static int GetSymHighlightMarker(_str &strKey)
{
    SymColour *symColor = getSymColour(strKey);
    return symColor->m_ScrollMarkerType;
}


/* D E F E R R E D  I N I T  S Y M  H I G H L I G H T */
/*----------------------------------------------------------------------------
    %%Function: DeferredInitSymHighlight
    %%Author: MarkSun
    %%Owner: MarkSun

   Start up our callback function.
   I do not know why deferring the init is good, I copied this from another
   macro.

----------------------------------------------------------------------------*/
static void DeferredInitSymHighlight()
{

    dbgsay("DeferredInitSymHighlight");
    if ( !pos( "-mdihide", _editor_cmdline, 1, 'i' ) )
    {
        dbgsay("DeferredInitSymHighlight 2");
        if ( s_markertypeSymHighlight < 0 )
        {
            s_markertypeSymHighlight = _MarkerTypeAlloc();
        }
        else
        {
            _StreamMarkerRemoveAllType( s_markertypeSymHighlight );
            RemoveAllSymHighlightMarkerTypes();
        }

        if ( s_timerSymHighlight >= 0 )
            _kill_timer( s_timerSymHighlight );

        dbgsay("DeferredInitSymHighlight3");
        _InitSymColour();
        if (!def_sym_highlight_persist_across_sessions)
        {
            DoClearAllSymHighlightStructs();
        }
        s_timerSymHighlight = _set_timer( def_sym_highlight_delay, SymHighlightCallback );
    }
}


/* D E F I N I T */
/*----------------------------------------------------------------------------
    %%Function: definit
    %%Author: MarkSun
    %%Owner: MarkSun

    Initialize this module.

----------------------------------------------------------------------------*/
definit()
{
    s_modtimeSymHighlight = -1;

    if (arg(1) != 'L')
    {
        s_timerSymHighlight = -1;
        s_markertypeSymHighlight = -1;
    }

    dbgsay("DEFINIT");
    _post_call( DeferredInitSymHighlight );
}

_exit_cleanup()
{
    int i;
    for (i=0; i < s_symColour._length(); i++)
    {
        if (s_symColour[i].m_ColourSymHighlight != -1)
        {
            _FreeColor(s_symColour[i].m_ColourSymHighlight);
        }
    }
}

_command void run_highlights()
{
    // The -modal option displays other windows while the dialog box
    // is displayed.
    show("-mdi Highlights");
}


